@@ -1,4 +1,6 @@
 import multiprocessing
+import warnings
+from math import ceil
 from math import floor
 from pathlib import Path
 from typing import List
@@ -10,6 +12,7 @@ from typing import cast
 
 import numpy as np
 import pandas as pd
+from matplotlib import pyplot as plt
 from scipy.linalg import solve  # type: ignore
 from scipy.optimize import minimize  # type: ignore
 from scipy.special import gammaln  # type: ignore
@@ -24,21 +27,21 @@ from pydeseq2.grid_search import grid_fit_shrink_beta
 
 
 def load_example_data(
-    modality: Literal["raw_counts", "clinical"] = "raw_counts",
+    modality: Literal["raw_counts", "metadata"] = "raw_counts",
     dataset: Literal["synthetic"] = "synthetic",
     debug: bool = False,
     debug_seed: int = 42,
 ) -> pd.DataFrame:
     """Load synthetic example data.
 
-    May load either clinical or rna-seq data. For now, this function may only return the
+    May load either metadata or rna-seq data. For now, this function may only return the
     synthetic data provided as part of this repo, but new datasets might be added in the
     future.
 
     Parameters
     ----------
     modality : str
-        Data modality. "raw_counts" or "clinical".
+        Data modality. "raw_counts" or "metadata".
 
     dataset : str
         The dataset for which to return gene expression data.
@@ -47,7 +50,8 @@ def load_example_data(
 
     debug : bool
         If true, subsample 10 samples and 100 genes at random.
-        (Note that the "synthetic" dataset is already 10 x 100.) (default: ``False``).
+        (Note that the "synthetic" dataset is already 10 features 100.)
+        (default: ``False``).
 
     debug_seed : int
         Seed for the debug mode. (default: ``42``).
@@ -57,9 +61,8 @@ def load_example_data(
     pandas.DataFrame
         Requested data modality.
     """
-
-    assert modality in ["raw_counts", "clinical"], (
-        "The modality argument must be one of the following: " "raw_counts, clinical"
+    assert modality in ["raw_counts", "metadata"], (
+        "The modality argument must be one of the following: " "raw_counts, metadata"
     )
 
     assert dataset in [
@@ -75,7 +78,7 @@ def load_example_data(
             # Cast the Paths to strings to have coherent types wrt to the url case (that
             # does not handle Paths), else mypy throws an error.
             path_to_data_counts = str(path_to_data / "test_counts.csv")
-            path_to_data_clinical = str(path_to_data / "test_clinical.csv")
+            path_to_data_metadata = str(path_to_data / "test_metadata.csv")
         else:
             # if the path does not exist (as is the case in RDT) load it from github
             url_to_data = (
@@ -83,7 +86,7 @@ def load_example_data(
                 "PyDESeq2/main/datasets/synthetic/"
             )
             path_to_data_counts = url_to_data + "/test_counts.csv"
-            path_to_data_clinical = url_to_data + "/test_clinical.csv"
+            path_to_data_metadata = url_to_data + "/test_metadata.csv"
 
         if modality == "raw_counts":
             df = pd.read_csv(
@@ -91,9 +94,9 @@ def load_example_data(
                 sep=",",
                 index_col=0,
             ).T
-        elif modality == "clinical":
+        elif modality == "metadata":
             df = pd.read_csv(
-                path_to_data_clinical,
+                path_to_data_metadata,
                 sep=",",
                 index_col=0,
             )
@@ -137,35 +140,39 @@ def test_valid_counts(counts: Union[pd.DataFrame, np.ndarray]) -> None:
 
 
 def build_design_matrix(
-    clinical_df: pd.DataFrame,
+    metadata: pd.DataFrame,
     design_factors: Union[str, List[str]] = "condition",
     ref_level: Optional[List[str]] = None,
+    continuous_factors: Optional[List[str]] = None,
     expanded: bool = False,
     intercept: bool = True,
 ) -> pd.DataFrame:
     """Build design_matrix matrix for DEA.
 
-    Only single factor, 2-level designs are currently supported.
     Unless specified, the reference factor is chosen alphabetically.
-    NOTE: For now, each factor should have exactly two levels.
 
     Parameters
     ----------
-    clinical_df : pandas.DataFrame
-        DataFrame containing clinical information.
+    metadata : pandas.DataFrame
+        DataFrame containing metadata information.
         Must be indexed by sample barcodes.
 
     design_factors : str or list
-        Name of the columns of clinical_df to be used as design_matrix variables.
+        Name of the columns of metadata to be used as design_matrix variables.
         (default: ``"condition"``).
 
-    ref_level : list or None
+    ref_level : dict or None
         An optional list of two strings of the form ``["factor", "ref_level"]``
         specifying the factor of interest and the desired reference level, e.g.
         ``["condition", "A"]``. (default: ``None``).
 
+    continuous_factors : list or None
+        An optional list of continuous (as opposed to categorical) factors. Any factor
+        not in ``continuous_factors`` will be considered categorical (default: ``None``).
+
     expanded : bool
-        If true, use one column per category. Else, use a single column.
+        If true, use one column per category. Else, use n-1 columns, for each n-level
+        categorical factor.
         (default: ``False``).
 
     intercept : bool
@@ -177,7 +184,6 @@ def build_design_matrix(
         A DataFrame with experiment design information (to split cohorts).
         Indexed by sample barcodes.
     """
-
     if isinstance(
         design_factors, str
     ):  # if there is a single factor, convert to singleton list
@@ -185,61 +191,118 @@ def build_design_matrix(
 
     for factor in design_factors:
         # Check that each factor has at least 2 levels
-        if len(np.unique(clinical_df[factor])) < 2:
+        if len(np.unique(metadata[factor])) < 2:
             raise ValueError(
                 f"Factors should take at least two values, but {factor} "
-                f"takes the single value '{np.unique(clinical_df[factor])}'."
+                f"takes the single value '{np.unique(metadata[factor])}'."
             )
 
-    design_matrix = pd.get_dummies(clinical_df[design_factors], drop_first=not expanded)
-
-    if ref_level is not None:
-        if len(ref_level) != 2:
-            raise KeyError("The reference level should contain 2 strings.")
-        if ref_level[1] not in clinical_df[ref_level[0]].values:
-            raise KeyError(
-                f"The clinical data should contain a '{ref_level[0]}' column"
-                f" with a '{ref_level[1]}' level."
-            )
-
-        # Check that the reference level is not in the matrix (if unexpanded design)
-        ref_level_name = "_".join(ref_level)
-        if (not expanded) and ref_level_name in design_matrix.columns:
-            # Remove the reference level and add one
-            factor_cols = [
-                col for col in design_matrix.columns if col.startswith(ref_level[0])
-            ]
-            missing_level = next(
-                level
-                for level in np.unique(clinical_df[ref_level[0]])
-                if f"{ref_level[0]}_{level}" not in design_matrix.columns
-            )
-            design_matrix[f"{ref_level[0]}_{missing_level}"] = 1 - design_matrix[
-                factor_cols
-            ].sum(1)
-            design_matrix.drop(ref_level_name, axis="columns", inplace=True)
-
-    if not expanded:
-        # Add reference level as column name suffix
-        for factor in design_factors:
-            if ref_level is None or factor != ref_level[0]:
-                # The reference is the unique level that is no longer there
-                ref = next(
-                    level
-                    for level in np.unique(clinical_df[factor])
-                    if f"{factor}_{level}" not in design_matrix.columns
+    # Check that level factors in the design don't contain underscores. If so, convert
+    # them to hyphens
+    warning_issued = False
+    for factor in design_factors:
+        if np.any(["_" in value for value in metadata[factor]]):
+            if not warning_issued:
+                warnings.warn(
+                    """Some factor levels in the design contain underscores ('_').
+                    They will be converted to hyphens ('-').""",
+                    UserWarning,
+                    stacklevel=2,
                 )
-            else:
-                # The reference level is given as an argument
-                ref = ref_level[1]
-            design_matrix.columns = [
-                f"{col}_vs_{ref}" if col.startswith(factor) else col
-                for col in design_matrix.columns
-            ]
+                warning_issued = True
+            metadata[factor] = metadata[factor].apply(lambda x: x.replace("_", "-"))
+
+    if continuous_factors is not None:
+        categorical_factors = [
+            factor for factor in design_factors if factor not in continuous_factors
+        ]
+    else:
+        categorical_factors = design_factors
+
+    # Check that there is at least one categorical factor
+    if len(categorical_factors) > 0:
+        design_matrix = pd.get_dummies(
+            metadata[categorical_factors], drop_first=not expanded
+        )
+
+        if ref_level is not None:
+            if len(ref_level) != 2:
+                raise KeyError("The reference level should contain 2 strings.")
+            if ref_level[1] not in metadata[ref_level[0]].values:
+                raise KeyError(
+                    f"The metadata data should contain a '{ref_level[0]}' column"
+                    f" with a '{ref_level[1]}' level."
+                )
+
+            # Check that the reference level is not in the matrix (if unexpanded design)
+            ref_level_name = "_".join(ref_level)
+            if (not expanded) and ref_level_name in design_matrix.columns:
+                # Remove the reference level and add one
+                factor_cols = [
+                    col for col in design_matrix.columns if col.startswith(ref_level[0])
+                ]
+                missing_level = next(
+                    level
+                    for level in np.unique(metadata[ref_level[0]])
+                    if f"{ref_level[0]}_{level}" not in design_matrix.columns
+                )
+                design_matrix[f"{ref_level[0]}_{missing_level}"] = 1 - design_matrix[
+                    factor_cols
+                ].sum(1)
+                design_matrix.drop(ref_level_name, axis="columns", inplace=True)
+
+        if not expanded:
+            # Add reference level as column name suffix
+            for factor in design_factors:
+                if ref_level is None or factor != ref_level[0]:
+                    # The reference is the unique level that is no longer there
+                    ref = next(
+                        level
+                        for level in np.unique(metadata[factor])
+                        if f"{factor}_{level}" not in design_matrix.columns
+                    )
+                else:
+                    # The reference level is given as an argument
+                    ref = ref_level[1]
+                design_matrix.columns = [
+                    f"{col}_vs_{ref}" if col.startswith(factor) else col
+                    for col in design_matrix.columns
+                ]
+    else:
+        # There is no categorical factor in the design
+        design_matrix = pd.DataFrame(index=metadata.index)
 
     if intercept:
         design_matrix.insert(0, "intercept", 1)
-    return design_matrix.astype("int")
+
+    # Convert categorical factors one-hot encodings to int
+    design_matrix = design_matrix.astype("int")
+
+    # Add continuous factors
+    if continuous_factors is not None:
+        for factor in continuous_factors:
+            # This factor should be numeric
+            design_matrix[factor] = pd.to_numeric(metadata[factor])
+    return design_matrix
+
+
+def replace_underscores(factors: List[str]):
+    """Replace all underscores from strings in a list by hyphens.
+
+    To be used on design factors to avoid bugs due to the reliance on `str.split("_")`
+    in parts of the code.
+
+    Parameters
+    ----------
+    factors : list
+        A list of strings which may contain underscores.
+
+    Returns
+    -------
+    list
+        A list of strings in which underscores were replaced by hyphens.
+    """
+    return [factor.replace("_", "-") for factor in factors]
 
 
 def dispersion_trend(
@@ -272,7 +335,7 @@ def dispersion_trend(
 def nb_nll(
     counts: np.ndarray, mu: np.ndarray, alpha: Union[float, np.ndarray]
 ) -> Union[float, np.ndarray]:
-    """Negative log-likelihood of a negative binomial of parameters ``mu`` and ``alpha``.
+    r"""Neg log-likelihood of a negative binomial of parameters ``mu`` and ``alpha``.
 
     Mathematically, if ``counts`` is a vector of counting entries :math:`y_i`
     then the likelihood of each entry :math:`y_i` to be drawn from a negative
@@ -322,7 +385,6 @@ def nb_nll(
     -----
     [1] https://en.wikipedia.org/wiki/Negative_binomial_distribution
     """
-
     n = len(counts)
     alpha_neg1 = 1 / alpha
     logbinom = gammaln(counts + alpha_neg1) - gammaln(counts + 1) - gammaln(alpha_neg1)
@@ -345,7 +407,7 @@ def nb_nll(
 
 
 def dnb_nll(counts: np.ndarray, mu: np.ndarray, alpha: float) -> float:
-    """Gradient of the negative log-likelihood of a negative binomial.
+    r"""Gradient of the negative log-likelihood of a negative binomial.
 
     Unvectorized.
 
@@ -366,7 +428,6 @@ def dnb_nll(counts: np.ndarray, mu: np.ndarray, alpha: float) -> float:
     float
         Derivative of negative log likelihood of NB w.r.t. :math:`\\alpha`.
     """
-
     alpha_neg1 = 1 / alpha
     ll_part = (
         alpha_neg1**2
@@ -451,7 +512,6 @@ def irls_solver(
         Whether IRLS or the optimizer converged. If not and if dimension allows it,
         perform grid search.
     """
-
     assert optimizer in ["BFGS", "L-BFGS-B"]
 
     num_vars = design_matrix.shape[1]
@@ -463,6 +523,9 @@ def irls_solver(
         y = np.log(counts / size_factors + 0.1)
         beta_init = solve(R, Q.T @ y)
         beta = beta_init
+    else:  # Initialise intercept with log base mean
+        beta = np.zeros(num_vars)
+        beta[0] = np.log(counts / size_factors).mean()
 
     dev = 1000.0
     dev_ratio = 1.0
@@ -595,7 +658,6 @@ def fit_alpha_mle(
     bool
         Whether L-BFGS-B converged. If not, dispersion is estimated using grid search.
     """
-
     assert optimizer in ["BFGS", "L-BFGS-B"]
 
     if prior_reg:
@@ -671,7 +733,7 @@ def trimmed_mean(x, trim: float = 0.1, **kwargs) -> Union[float, np.ndarray]:
 
     Parameters
     ----------
-    x : ndarray
+    features : ndarray
         Data whose mean to compute.
 
     trim : float
@@ -685,7 +747,6 @@ def trimmed_mean(x, trim: float = 0.1, **kwargs) -> Union[float, np.ndarray]:
     float or ndarray :
         Trimmed mean.
     """
-
     assert trim <= 0.5
     if "axis" in kwargs:
         axis = kwargs.get("axis")
@@ -720,7 +781,6 @@ def trimmed_cell_variance(counts: pd.DataFrame, cells: pd.Series) -> pd.Series:
     pandas.Series :
         Gene-wise trimmed variance estimate.
     """
-
     # how much to trim at different n
     trimratio = (1 / 3, 1 / 4, 1 / 8)
     # returns an index for the vector above for three sample size bins
@@ -757,7 +817,7 @@ def trimmed_variance(
 
     Parameters
     ----------
-    x : ndarray
+    features : ndarray
         Data whose trimmed variance to compute.
 
     trim : float
@@ -771,7 +831,6 @@ def trimmed_variance(
     float or ndarray
         Trimmed variances.
     """
-
     rm = trimmed_mean(x, trim=trim, axis=axis)
     sqerror = (x - rm) ** 2
     # scale due to trimming of large squares
@@ -807,7 +866,6 @@ def fit_lin_mu(
     ndarray
         Estimated mean.
     """
-
     reg = LinearRegression(fit_intercept=False)
     reg.fit(design_matrix, counts / size_factors)
     mu_hat = size_factors * reg.predict(design_matrix)
@@ -822,6 +880,8 @@ def wald_test(
     mu: float,
     ridge_factor: np.ndarray,
     contrast: np.ndarray,
+    lfc_null: float,
+    alt_hypothesis: Optional[Literal["greaterAbs", "lessAbs", "greater", "less"]],
 ) -> Tuple[float, float, float]:
     """Run Wald test for differential expression.
 
@@ -848,6 +908,12 @@ def wald_test(
     contrast : ndarray
         Vector encoding the contrast that is being tested.
 
+    lfc_null : float
+        The (log2) log fold change under the null hypothesis.
+
+    alt_hypothesis : str or None
+        The alternative hypothesis for computing wald p-values.
+
     Returns
     -------
     wald_p_value : float
@@ -859,7 +925,6 @@ def wald_test(
     wald_se : float
         Standard error of the Wald statistic.
     """
-
     # Build covariance matrix estimator
     W = np.diag(mu / (1 + mu * disp))
     M = design_matrix.T @ W @ design_matrix
@@ -867,26 +932,55 @@ def wald_test(
     Hc = H @ contrast
     # Evaluate standard error and Wald statistic
     wald_se: float = np.sqrt(Hc.T @ M @ Hc)
-    wald_statistic: float = contrast @ lfc / wald_se
-    wald_p_value = 2 * norm.sf(np.abs(wald_statistic))
+
+    def greater(lfc_null):
+        stat = contrast @ np.fmax((lfc - lfc_null) / wald_se, 0)
+        pval = norm.sf(stat)
+        return stat, pval
+
+    def less(lfc_null):
+        stat = contrast @ np.fmin((lfc - lfc_null) / wald_se, 0)
+        pval = norm.sf(np.abs(stat))
+        return stat, pval
+
+    def greater_abs(lfc_null):
+        stat = contrast @ (np.sign(lfc) * np.fmax((np.abs(lfc) - lfc_null) / wald_se, 0))
+        pval = 2 * norm.sf(np.abs(stat))  # Only case where the test is two-tailed
+        return stat, pval
+
+    def less_abs(lfc_null):
+        stat_above, pval_above = greater(-abs(lfc_null))
+        stat_below, pval_below = less(abs(lfc_null))
+        return min(stat_above, stat_below, key=abs), max(pval_above, pval_below)
+
+    wald_statistic: float
+    wald_p_value: float
+    if alt_hypothesis:
+        wald_statistic, wald_p_value = {
+            "greaterAbs": greater_abs(lfc_null),
+            "lessAbs": less_abs(lfc_null),
+            "greater": greater(lfc_null),
+            "less": less(lfc_null),
+        }[alt_hypothesis]
+    else:
+        wald_statistic = contrast @ (lfc - lfc_null) / wald_se
+        wald_p_value = 2 * norm.sf(np.abs(wald_statistic))
+
     return wald_p_value, wald_statistic, wald_se
 
 
 def fit_rough_dispersions(
-    counts: np.ndarray, size_factors: np.ndarray, design_matrix: pd.DataFrame
+    normed_counts: np.ndarray, design_matrix: pd.DataFrame
 ) -> np.ndarray:
-    """ "Rough dispersion" estimates from linear model, as per the R code.
+    """Rough dispersion estimates from linear model, as per the R code.
 
     Used as initial estimates in :meth:`DeseqDataSet.fit_genewise_dispersions()
     <pydeseq2.dds.DeseqDataSet.fit_genewise_dispersions>`.
 
     Parameters
     ----------
-    counts : ndarray
-        Raw counts. One column per gene, one row per sample.
-
-    size_factors : ndarray
-        DESeq2 normalization factors.
+    normed_counts : ndarray
+        Array of deseq2-normalized read counts. Rows: samples, columns: genes.
 
     design_matrix : pandas.DataFrame
         A DataFrame with experiment design information (to split cohorts).
@@ -897,13 +991,15 @@ def fit_rough_dispersions(
     ndarray
         Estimated dispersion parameter for each gene.
     """
-
     num_samples, num_vars = design_matrix.shape
-
-    normed_counts = counts / size_factors[:, None]
-
-    # Exclude genes with all zeroes
-    normed_counts = normed_counts[:, ~(normed_counts == 0).all(axis=0)]
+    # This method is only possible when num_samples > num_vars.
+    # If this is not the case, throw an error.
+    if num_samples == num_vars:
+        raise ValueError(
+            "The number of samples and the number of design variables are "
+            "equal, i.e., there are no replicates to estimate the "
+            "dispersion. Please use a design with fewer variables."
+        )
 
     reg = LinearRegression(fit_intercept=False)
     reg.fit(design_matrix, normed_counts)
@@ -915,7 +1011,9 @@ def fit_rough_dispersions(
     return np.maximum(alpha_rde, 0)
 
 
-def fit_moments_dispersions(counts: np.ndarray, size_factors: np.ndarray) -> np.ndarray:
+def fit_moments_dispersions(
+    normed_counts: np.ndarray, size_factors: np.ndarray
+) -> np.ndarray:
     """Dispersion estimates based on moments, as per the R code.
 
     Used as initial estimates in :meth:`DeseqDataSet.fit_genewise_dispersions()
@@ -923,8 +1021,8 @@ def fit_moments_dispersions(counts: np.ndarray, size_factors: np.ndarray) -> np.
 
     Parameters
     ----------
-    counts : ndarray
-        Raw counts. One column per gene, one row per sample.
+    normed_counts : ndarray
+        Array of deseq2-normalized read counts. Rows: samples, columns: genes.
 
     size_factors : ndarray
         DESeq2 normalization factors.
@@ -934,8 +1032,6 @@ def fit_moments_dispersions(counts: np.ndarray, size_factors: np.ndarray) -> np.
     ndarray
         Estimated dispersion parameter for each gene.
     """
-
-    normed_counts = counts / size_factors[:, None]
     # Exclude genes with all zeroes
     normed_counts = normed_counts[:, ~(normed_counts == 0).all(axis=0)]
     # mean inverse size factor
@@ -947,6 +1043,32 @@ def fit_moments_dispersions(counts: np.ndarray, size_factors: np.ndarray) -> np.
     return np.nan_to_num((sigma - s_mean_inv * mu) / mu**2)
 
 
+def n_or_more_replicates(design_matrix: pd.DataFrame, min_replicates: int) -> pd.Series:
+    """
+    Return a  series indicating whether samples have a minimum number of replicates.
+
+    Checks whether each sample has at least ``min_replicates`` replicates, based on its
+    combination of design factors.
+
+    Parameters
+    ----------
+    design_matrix : pandas.DataFrame
+        A DataFrame with experiment design information (to split cohorts).
+    min_replicates : int
+        The minimum number of replicates to have to pass the threshold.
+
+    Returns
+    -------
+    pandas.Series
+        A boolean series indicating whether each sample has at least ``min_replicates``
+        replicates.
+    """
+    n_or_more = design_matrix.value_counts() >= min_replicates
+    replaceable = n_or_more[pd.MultiIndex.from_frame(design_matrix)]
+    replaceable.index = design_matrix.index
+    return replaceable
+
+
 def robust_method_of_moments_disp(
     normed_counts: pd.DataFrame, design_matrix: pd.DataFrame
 ) -> pd.Series:
@@ -969,12 +1091,21 @@ def robust_method_of_moments_disp(
         Trimmed method of moment dispersion estimates.
         Used for outlier detection based on Cook's distance.
     """
-
     # if there are 3 or more replicates in any cell
-    three_or_more = design_matrix[design_matrix.columns[-1]].value_counts() >= 3
+    three_or_more = n_or_more_replicates(design_matrix, 3)
     if three_or_more.any():
-        v = trimmed_cell_variance(normed_counts, design_matrix.iloc[:, -1].astype(int))
-        # last argument is non-intercept.
+        # 1 - group rows by unique combinations of design factors
+        # 2 - keep only groups with 3 or more replicates
+        # 3 - filter the counts matrix to only keep rows in those groups
+        filtered_counts = normed_counts.loc[three_or_more, :]
+        filtered_design = design_matrix.loc[three_or_more, :]
+        cell_id = pd.Series(
+            filtered_design.groupby(
+                filtered_design.columns.values.tolist()
+            ).grouper.group_info[0],
+            index=filtered_design.index,
+        )
+        v = trimmed_cell_variance(filtered_counts, cell_id)
     else:
         v = pd.Series(
             trimmed_variance(normed_counts.values), index=normed_counts.columns
@@ -1004,7 +1135,6 @@ def get_num_processes(n_cpus: Optional[int] = None) -> int:
     int
         Number of processes to spawn.
     """
-
     if n_cpus is None:
         try:
             n_processes = multiprocessing.cpu_count()
@@ -1068,7 +1198,6 @@ def nbinomGLM(
     converged: bool
         Whether L-BFGS-B converged.
     """
-
     num_vars = design_matrix.shape[-1]
 
     shrink_mask = np.zeros(num_vars)
@@ -1215,7 +1344,6 @@ def nbinomFn(
     float
         Sum of the NB negative likelihood and apeGLM prior.
     """
-
     num_vars = design_matrix.shape[-1]
 
     shrink_mask = np.zeros(num_vars)
@@ -1242,7 +1370,7 @@ def mean_absolute_deviation(x: np.ndarray) -> float:
 
     Parameters
     ----------
-    x : ndarray
+    features : ndarray
         1D array whose MAD to compute.
 
     Returns
@@ -1252,3 +1380,215 @@ def mean_absolute_deviation(x: np.ndarray) -> float:
     """
     center = np.median(x)
     return np.median(np.abs(x - center)) / norm.ppf(0.75)
+
+
+def make_scatter(
+    disps: list,
+    legend_labels: list,
+    x_val: np.array,
+    log: bool = True,
+    save_path: Optional[str] = None,
+    **kwargs,
+) -> None:
+    """
+    Create a scatter plot using matplotlib.
+
+    Used in :meth:`pydeseq2.dds.DeseqDataSet.plot_dispersions()`.
+
+    Parameters
+    ----------
+    disps : list
+        List of ndarrays to plot.
+
+    legend_labels : list
+        List of strings that correspond to plotted targets values for legend.
+
+    x_val : ndarray
+        1D array to plot (example: ``dds.varm['_normed_means']``).
+
+    log : bool
+        Whether or not to log scale features and targets axes (``default=True``).
+
+    save_path : str or None
+        The path where to save the plot. If left None, the plot won't be saved
+        (``default=None``).
+
+    **kwargs
+        Matplotlib keyword arguments for the scatter plot.
+    """
+    # Adding more colors if plotting more than 3 traces
+    if len(disps) == 3:
+        colors = "kbr"
+    else:
+        colors = "kbrcmyg"
+
+    # Standardizing font; init plot
+    plt.rcParams.update({"font.size": 10})
+    fig, ax = plt.subplots(dpi=600)
+
+    # log scale axes
+    if log is True:
+        plt.yscale("log")
+        plt.xscale("log")
+
+    # scale axes according to data spread
+    ax.set_adjustable("datalim")
+
+    # Set default alpha and s parameters, if not already specified
+    kwargs.setdefault("alpha", 0.5)
+    kwargs.setdefault("s", 0.6)
+
+    # create scatter plot per trace
+    for disp, color in list(zip(disps, colors)):
+        plt.scatter(x=x_val, y=disp, c=color, **kwargs)
+
+    # label legend + axes
+    plt.legend(legend_labels, loc="best")
+    plt.xlabel("mean of normalized counts")
+    plt.ylabel("dispersion")
+    plt.tight_layout()
+    if save_path is not None:
+        plt.savefig(save_path, bbox_inches="tight")
+    plt.show()
+
+
+def make_MA_plot(
+    results_df: pd.DataFrame,
+    padj_thresh: float = 0.05,
+    log: bool = True,
+    save_path: Optional[str] = None,
+    lfc_null: float = 0,
+    alt_hypothesis: Optional[Literal["greaterAbs", "lessAbs", "greater", "less"]] = None,
+    **kwargs,
+) -> None:
+    """
+    Create an log ratio (M)-average (A) plot using matplotlib.
+
+    Useful for looking at log fold-change versus mean expression
+    between two groups/samples/etc.
+    Uses matplotlib to emulate make_MA() function in DESeq2 in R.
+
+    Parameters
+    ----------
+    results_df : pd.DataFrame
+        Resultant dataframe after running DeseqStats() and .summary().
+
+    padj_thresh : float
+        P-value threshold to subset scatterplot colors on.
+
+    log : bool
+        Whether or not to log scale features and targets axes (``default=True``).
+
+    save_path : str or None
+        The path where to save the plot. If left None, the plot won't be saved
+        (``default=None``).
+
+    lfc_null : float
+        The (log2) log fold change under the null hypothesis. (default: ``0``).
+
+    alt_hypothesis : str or None
+        The alternative hypothesis for computing wald p-values. (default: ``None``).
+
+    **kwargs
+        Matplotlib keyword arguments for the scatter plot.
+    """
+    colors = results_df["padj"].apply(lambda x: "darkred" if x < padj_thresh else "gray")
+
+    fig, ax = plt.subplots(dpi=600)
+
+    # Set default alpha and s parameters, if not already specified
+    kwargs.setdefault("alpha", 0.5)
+    kwargs.setdefault("s", 0.2)
+
+    plt.scatter(
+        x=results_df["baseMean"],
+        y=results_df["log2FoldChange"],
+        c=colors,
+        **kwargs,
+    )
+
+    ax.set_adjustable("datalim")
+
+    if log is True:
+        plt.xscale("log")
+
+    plt.xlabel("mean of normalized counts")
+    plt.ylabel("log2 fold change")
+
+    plt.axhline(lfc_null, color="red", alpha=0.5, linestyle="--", zorder=3)
+    if alt_hypothesis and alt_hypothesis in ["greaterAbs", "lessAbs"]:
+        plt.axhline(-lfc_null, color="red", alpha=0.5, linestyle="--", zorder=3)
+    plt.tight_layout()
+
+    if save_path is not None:
+        plt.savefig(save_path, bbox_inches="tight")
+
+
+# Authors: Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>
+#
+# License: BSD (3-clause)
+
+# Adapted from https://gist.github.com/agramfort/850437
+
+
+def lowess(
+    features: np.ndarray, targets: np.ndarray, frac: float = 2.0 / 3.0, iter: int = 3
+):
+    """Run lowess smoothing: Robust locally weighted regression.
+
+    The lowess function fits a nonparametric regression curve to a scatterplot.
+    The arrays features and targets contain an equal number of elements; each pair
+    (features[i], targets[i]) defines a data point in the scatterplot. The function
+    returns the estimated (smooth) values of targets.
+    The smoothing span is given by frac. A larger value for frac will result in a
+    smoother curve. The number of robustifying iterations is given by iter. The
+    function will run faster with a smaller number of iterations.
+
+    Parameters
+    ----------
+    features : ndarray
+        A 1D array of data points.
+    targets : ndarray
+        A 1D array of target values (with the same shape as features).
+    frac : float
+        The fraction of the data used when estimating each y-value. (default: ``2/3``).
+    iter : int
+        The number of robustifying iterations. (default: ``3``).
+
+    Returns
+    -------
+    ndarray
+        Estimated (smooth) values of targets.
+    """
+    n = len(features)
+    r = int(ceil(frac * n))
+    h = np.maximum(
+        np.array([np.sort(np.abs(features - features[i]))[r] for i in range(n)]), 1e-12
+    )
+    w = np.clip(
+        np.abs(np.nan_to_num((features[:, None] - features[None, :]) / h)), 0.0, 1.0
+    )
+    w = (1 - w**3) ** 3
+    yest = np.zeros(n)
+    delta = np.ones(n)
+    for _ in range(iter):
+        for i in range(n):
+            weights = delta * w[:, i]
+            b = np.array(
+                [np.sum(weights * targets), np.sum(weights * targets * features)]
+            )
+            A = np.array(
+                [
+                    [np.sum(weights), np.sum(weights * features)],
+                    [np.sum(weights * features), np.sum(weights * features * features)],
+                ]
+            )
+            beta = np.linalg.lstsq(A, b)[0]
+            yest[i] = beta[0] + beta[1] * features[i]
+
+        residuals = targets - yest
+        s = np.median(np.abs(residuals))
+        delta = np.clip(residuals / (6.0 * s), -1, 1)
+        delta = (1 - delta**2) ** 2
+
+    return yest
