@@ -4,41 +4,49 @@ from pathlib import Path
 import anndata as ad
 import numpy as np
 import pandas as pd
+import pytest
 
 import tests
 from pydeseq2.dds import DeseqDataSet
 from pydeseq2.ds import DeseqStats
+from pydeseq2.preprocessing import deseq2_norm
+from pydeseq2.preprocessing import deseq2_norm_fit
+from pydeseq2.preprocessing import deseq2_norm_transform
 from pydeseq2.utils import load_example_data
 
 # Single-factor tests
 
 
-def test_deseq(tol=0.02):
+@pytest.fixture
+def counts_df():
+    return load_example_data(
+        modality="raw_counts",
+        dataset="synthetic",
+        debug=False,
+    )
+
+
+@pytest.fixture
+def metadata():
+    return load_example_data(
+        modality="metadata",
+        dataset="synthetic",
+        debug=False,
+    )
+
+
+def test_deseq(counts_df, metadata, tol=0.02):
     """Test that the outputs of the DESeq2 function match those of the original R
     package, up to a tolerance in relative error.
     """
 
     test_path = str(Path(os.path.realpath(tests.__file__)).parent.resolve())
 
-    counts_df = load_example_data(
-        modality="raw_counts",
-        dataset="synthetic",
-        debug=False,
-    )
-
-    clinical_df = load_example_data(
-        modality="clinical",
-        dataset="synthetic",
-        debug=False,
-    )
-
     r_res = pd.read_csv(
         os.path.join(test_path, "data/single_factor/r_test_res.csv"), index_col=0
     )
 
-    dds = DeseqDataSet(
-        counts=counts_df, clinical=clinical_df, design_factors="condition"
-    )
+    dds = DeseqDataSet(counts=counts_df, metadata=metadata, design_factors="condition")
     dds.deseq2()
 
     res = DeseqStats(dds)
@@ -57,7 +65,58 @@ def test_deseq(tol=0.02):
     assert (abs(r_res.padj - res_df.padj) / r_res.padj).max() < tol
 
 
-def test_deseq_no_refit_cooks(tol=0.02):
+@pytest.mark.parametrize("alt_hypothesis", ["lessAbs", "greaterAbs", "less", "greater"])
+def test_alt_hypothesis(alt_hypothesis, counts_df, metadata, tol=0.02):
+    """Test that the outputs of the DESeq2 function match those of the original R
+    package, up to a tolerance in relative error, with the alternative hypothesis test.
+    """
+
+    test_path = str(Path(os.path.realpath(tests.__file__)).parent.resolve())
+
+    r_res = pd.read_csv(
+        os.path.join(test_path, f"data/single_factor/r_test_res_{alt_hypothesis}.csv"),
+        index_col=0,
+    )
+
+    dds = DeseqDataSet(
+        counts=counts_df,
+        metadata=metadata,
+        design_factors="condition",
+    )
+    dds.deseq2()
+
+    res = DeseqStats(
+        dds,
+        lfc_null=-0.5 if alt_hypothesis == "less" else 0.5,
+        alt_hypothesis=alt_hypothesis,
+    )
+    res.summary()
+    res_df = res.results_df
+
+    # check that the same p-values are NaN
+    assert (res_df.pvalue.isna() == r_res.pvalue.isna()).all()
+    assert (res_df.padj.isna() == r_res.padj.isna()).all()
+
+    # Check that the same LFC and Wald statistics are found (up to tol)
+    assert (
+        abs(r_res.log2FoldChange - res_df.log2FoldChange) / abs(r_res.log2FoldChange)
+    ).max() < tol
+    if alt_hypothesis == "lessAbs":
+        res_df.stat = res_df.stat.abs()
+    assert (abs(r_res.stat - res_df.stat) / abs(r_res.stat)).max() < tol
+    # Check for the same pvalue and padj where stat != 0
+    assert (
+        abs(r_res.pvalue[r_res.stat != 0] - res_df.pvalue[res_df.stat != 0])
+        / r_res.pvalue[r_res.stat != 0]
+    ).max() < tol
+    # Not passing
+    # assert (
+    #     abs(r_res.padj[r_res.stat != 0] - res_df.padj[res_df.stat != 0])
+    #     / r_res.padj[r_res.stat != 0]
+    # ).max() < tol
+
+
+def test_deseq_no_refit_cooks(counts_df, metadata, tol=0.02):
     """Test that the outputs of the DESeq2 function *without cooks refit*
     match those of the original R package, up to a tolerance in relative error.
     Note: this is just to check that the workflow runs bug-free, as we expect no outliers
@@ -66,25 +125,13 @@ def test_deseq_no_refit_cooks(tol=0.02):
 
     test_path = str(Path(os.path.realpath(tests.__file__)).parent.resolve())
 
-    counts_df = load_example_data(
-        modality="raw_counts",
-        dataset="synthetic",
-        debug=False,
-    )
-
-    clinical_df = load_example_data(
-        modality="clinical",
-        dataset="synthetic",
-        debug=False,
-    )
-
     r_res = pd.read_csv(
         os.path.join(test_path, "data/single_factor/r_test_res.csv"), index_col=0
     )
 
     dds = DeseqDataSet(
         counts=counts_df,
-        clinical=clinical_df,
+        metadata=metadata,
         design_factors="condition",
         refit_cooks=False,
     )
@@ -106,7 +153,7 @@ def test_deseq_no_refit_cooks(tol=0.02):
     assert (abs(r_res.padj - res_df.padj) / r_res.padj).max() < tol
 
 
-def test_lfc_shrinkage(tol=0.02):
+def test_lfc_shrinkage(counts_df, metadata, tol=0.02):
     """Test that the outputs of the lfc_shrink function match those of the original
     R package (starting from the same inputs), up to a tolerance in relative error.
     """
@@ -120,18 +167,6 @@ def test_lfc_shrinkage(tol=0.02):
         index_col=0,
     )
 
-    counts_df = load_example_data(
-        modality="raw_counts",
-        dataset="synthetic",
-        debug=False,
-    )
-
-    clinical_df = load_example_data(
-        modality="clinical",
-        dataset="synthetic",
-        debug=False,
-    )
-
     r_size_factors = pd.read_csv(
         os.path.join(test_path, "data/single_factor/r_test_size_factors.csv"),
         index_col=0,
@@ -141,9 +176,7 @@ def test_lfc_shrinkage(tol=0.02):
         os.path.join(test_path, "data/single_factor/r_test_dispersions.csv"), index_col=0
     ).squeeze()
 
-    dds = DeseqDataSet(
-        counts=counts_df, clinical=clinical_df, design_factors="condition"
-    )
+    dds = DeseqDataSet(counts=counts_df, metadata=metadata, design_factors="condition")
     dds.deseq2()
     dds.obsm["size_factors"] = r_size_factors.values
     dds.varm["dispersions"] = r_dispersions.values
@@ -162,33 +195,20 @@ def test_lfc_shrinkage(tol=0.02):
     ).max() < tol
 
 
-def test_iterative_size_factors(tol=0.02):
+def test_iterative_size_factors(counts_df, metadata, tol=0.02):
     """Test that the outputs of the iterative size factor method match those of the
     original R package (starting from the same inputs), up to a tolerance in relative
     error.
     """
 
     test_path = str(Path(os.path.realpath(tests.__file__)).parent.resolve())
-    counts_df = load_example_data(
-        modality="raw_counts",
-        dataset="synthetic",
-        debug=False,
-    )
-
-    clinical_df = load_example_data(
-        modality="clinical",
-        dataset="synthetic",
-        debug=False,
-    )
 
     r_size_factors = pd.read_csv(
         os.path.join(test_path, "data/single_factor/r_iterative_size_factors.csv"),
         index_col=0,
     ).squeeze()
 
-    dds = DeseqDataSet(
-        counts=counts_df, clinical=clinical_df, design_factors="condition"
-    )
+    dds = DeseqDataSet(counts=counts_df, metadata=metadata, design_factors="condition")
     dds._fit_iterate_size_factors()
 
     # Check that the same LFC are found (up to tol)
@@ -198,35 +218,37 @@ def test_iterative_size_factors(tol=0.02):
 
 
 # Multi-factor tests
-def test_multifactor_deseq(tol=0.02):
+@pytest.mark.parametrize("with_outliers", [True, False])
+def test_multifactor_deseq(counts_df, metadata, with_outliers, tol=0.04):
     """Test that the outputs of the DESeq2 function match those of the original R
     package, up to a tolerance in relative error.
     """
 
     test_path = str(Path(os.path.realpath(tests.__file__)).parent.resolve())
 
-    counts_df = load_example_data(
-        modality="raw_counts",
-        dataset="synthetic",
-        debug=False,
-    )
+    if with_outliers:
+        r_res = pd.read_csv(
+            os.path.join(test_path, "data/multi_factor/r_test_res_outliers.csv"),
+            index_col=0,
+        )
+    else:
+        r_res = pd.read_csv(
+            os.path.join(test_path, "data/multi_factor/r_test_res.csv"),
+            index_col=0,
+        )
 
-    clinical_df = load_example_data(
-        modality="clinical",
-        dataset="synthetic",
-        debug=False,
-    )
-
-    r_res = pd.read_csv(
-        os.path.join(test_path, "data/multi_factor/r_test_res.csv"), index_col=0
-    )
+    if with_outliers:
+        # Introduce outlier counts and a condition level with a single replicate
+        counts_df.loc["sample1", "gene1"] = 2000
+        counts_df.loc["sample11", "gene7"] = 1000
+        metadata.loc["sample1", "condition"] = "C"
 
     dds = DeseqDataSet(
-        counts=counts_df, clinical=clinical_df, design_factors=["group", "condition"]
+        counts=counts_df, metadata=metadata, design_factors=["group", "condition"]
     )
     dds.deseq2()
 
-    res = DeseqStats(dds)
+    res = DeseqStats(dds, contrast=["condition", "B", "A"])
     res.summary()
     res_df = res.results_df
 
@@ -242,7 +264,7 @@ def test_multifactor_deseq(tol=0.02):
     assert (abs(r_res.padj - res_df.padj) / r_res.padj).max() < tol
 
 
-def test_multifactor_lfc_shrinkage(tol=0.02):
+def test_multifactor_lfc_shrinkage(counts_df, metadata, tol=0.02):
     """Test that the outputs of the lfc_shrink function match those of the original
     R package (starting from the same inputs), up to a tolerance in relative error.
     """
@@ -256,18 +278,6 @@ def test_multifactor_lfc_shrinkage(tol=0.02):
         index_col=0,
     )
 
-    counts_df = load_example_data(
-        modality="raw_counts",
-        dataset="synthetic",
-        debug=False,
-    )
-
-    clinical_df = load_example_data(
-        modality="clinical",
-        dataset="synthetic",
-        debug=False,
-    )
-
     r_size_factors = pd.read_csv(
         os.path.join(test_path, "data/multi_factor/r_test_size_factors.csv"),
         index_col=0,
@@ -278,7 +288,7 @@ def test_multifactor_lfc_shrinkage(tol=0.02):
     ).squeeze()
 
     dds = DeseqDataSet(
-        counts=counts_df, clinical=clinical_df, design_factors=["group", "condition"]
+        counts=counts_df, metadata=metadata, design_factors=["group", "condition"]
     )
     dds.deseq2()
     dds.obsm["size_factors"] = r_size_factors.values
@@ -298,26 +308,132 @@ def test_multifactor_lfc_shrinkage(tol=0.02):
     ).max() < tol
 
 
-def test_contrast():
+# Continuous tests
+@pytest.mark.parametrize("with_outliers", [True, False])
+def test_continuous_deseq(
+    with_outliers,
+    tol=0.04,
+):
+    """Test that the outputs of the DESeq2 function match those of the original R
+    package, up to a tolerance in relative error, with a continuous factor.
+    """
+
+    test_path = str(Path(os.path.realpath(tests.__file__)).parent.resolve())
+
+    counts_df = pd.read_csv(
+        os.path.join(test_path, "data/continuous/test_counts.csv"), index_col=0
+    ).T
+
+    metadata = pd.read_csv(
+        os.path.join(test_path, "data/continuous/test_metadata.csv"), index_col=0
+    )
+
+    # Load R results
+    if with_outliers:
+        r_res = pd.read_csv(
+            os.path.join(test_path, "data/continuous/r_test_res_outliers.csv"),
+            index_col=0,
+        )
+    else:
+        r_res = pd.read_csv(
+            os.path.join(test_path, "data/continuous/r_test_res.csv"), index_col=0
+        )
+
+    if with_outliers:
+        # Introduce outlier counts and a condition level with a single replicate
+        counts_df.loc["sample1", "gene1"] = 2000
+        counts_df.loc["sample11", "gene7"] = 1000
+        metadata.loc["sample1", "condition"] = "C"
+
+    dds = DeseqDataSet(
+        counts=counts_df,
+        metadata=metadata,
+        design_factors=["group", "condition", "measurement"],
+        continuous_factors=["measurement"],
+    )
+    dds.deseq2()
+
+    res = DeseqStats(dds)
+    res.summary()
+    res_df = res.results_df
+
+    # check that the same p-values are NaN
+    assert (res_df.pvalue.isna() == r_res.pvalue.isna()).all()
+    assert (res_df.padj.isna() == r_res.padj.isna()).all()
+
+    # Check that the same LFC, p-values and adjusted p-values are found (up to tol)
+    assert (
+        abs(r_res.log2FoldChange - res_df.log2FoldChange) / abs(r_res.log2FoldChange)
+    ).max() < tol
+    assert (abs(r_res.pvalue - res_df.pvalue) / r_res.pvalue).max() < tol
+    assert (abs(r_res.padj - res_df.padj) / r_res.padj).max() < tol
+
+
+def test_continuous_lfc_shrinkage(tol=0.02):
+    """Test that the outputs of the lfc_shrink function match those of the original
+    R package (starting from the same inputs), up to a tolerance in relative error.
+    """
+
+    test_path = str(Path(os.path.realpath(tests.__file__)).parent.resolve())
+
+    r_res = pd.read_csv(
+        os.path.join(test_path, "data/continuous/r_test_res.csv"), index_col=0
+    )
+    r_shrunk_res = pd.read_csv(
+        os.path.join(test_path, "data/continuous/r_test_lfc_shrink_res.csv"),
+        index_col=0,
+    )
+
+    counts_df = pd.read_csv(
+        os.path.join(test_path, "data/continuous/test_counts.csv"), index_col=0
+    ).T
+
+    metadata = pd.read_csv(
+        os.path.join(test_path, "data/continuous/test_metadata.csv"), index_col=0
+    )
+
+    r_size_factors = pd.read_csv(
+        os.path.join(test_path, "data/continuous/r_test_size_factors.csv"),
+        index_col=0,
+    ).squeeze()
+
+    r_dispersions = pd.read_csv(
+        os.path.join(test_path, "data/continuous/r_test_dispersions.csv"), index_col=0
+    ).squeeze()
+
+    dds = DeseqDataSet(
+        counts=counts_df,
+        metadata=metadata,
+        design_factors=["group", "condition", "measurement"],
+        continuous_factors=["measurement"],
+    )
+
+    dds.deseq2()
+    dds.obsm["size_factors"] = r_size_factors.values
+    dds.varm["dispersions"] = r_dispersions.values
+    dds.varm["LFC"].iloc[:, 1] = r_res.log2FoldChange.values * np.log(2)
+
+    res = DeseqStats(dds)
+    res.summary()
+    res.SE = r_res.lfcSE * np.log(2)
+    res.lfc_shrink(coeff="measurement")
+    shrunk_res = res.results_df
+
+    # Check that the same LFC found (up to tol)
+    assert (
+        abs(r_shrunk_res.log2FoldChange - shrunk_res.log2FoldChange)
+        / abs(r_shrunk_res.log2FoldChange)
+    ).max() < tol
+
+
+def test_contrast(counts_df, metadata):
     """
     Check that the contrasts ['condition', 'B', 'A'] and ['condition', 'A', 'B'] give
     coherent results (change of sign in LFCs and Wald stats, same (adjusted p-values).
     """
 
-    counts_df = load_example_data(
-        modality="raw_counts",
-        dataset="synthetic",
-        debug=False,
-    )
-
-    clinical_df = load_example_data(
-        modality="clinical",
-        dataset="synthetic",
-        debug=False,
-    )
-
     dds = DeseqDataSet(
-        counts=counts_df, clinical=clinical_df, design_factors=["group", "condition"]
+        counts=counts_df, metadata=metadata, design_factors=["group", "condition"]
     )
     dds.deseq2()
 
@@ -345,28 +461,15 @@ def test_contrast():
     )
 
 
-def test_anndata_init(tol=0.02):
+def test_anndata_init(counts_df, metadata, tol=0.02):
     """
     Test initializing dds with an AnnData object that already has filled in fields,
     including with the same names as those used by pydeseq2.
     """
     np.random.seed(42)
 
-    # Load data
-    counts_df = load_example_data(
-        modality="raw_counts",
-        dataset="synthetic",
-        debug=False,
-    )
-
-    clinical_df = load_example_data(
-        modality="clinical",
-        dataset="synthetic",
-        debug=False,
-    )
-
     # Make an anndata object
-    adata = ad.AnnData(X=counts_df, obs=clinical_df, dtype=int)
+    adata = ad.AnnData(X=counts_df.astype(int), obs=metadata)
 
     # Put some dummy data in unused fields
     adata.obsm["dummy_metadata"] = np.random.choice(2, adata.n_obs)
@@ -402,24 +505,10 @@ def test_anndata_init(tol=0.02):
     assert (abs(r_res.padj - res_df.padj) / r_res.padj).max() < tol
 
 
-def test_vst(tol=0.02):
+def test_vst(counts_df, metadata, tol=0.02):
     """
     Test the output of VST compared with DESeq2.
     """
-
-    # Load data
-    counts_df = load_example_data(
-        modality="raw_counts",
-        dataset="synthetic",
-        debug=False,
-    )
-
-    clinical_df = load_example_data(
-        modality="clinical",
-        dataset="synthetic",
-        debug=False,
-    )
-
     # Load R data
     test_path = str(Path(os.path.realpath(tests.__file__)).parent.resolve())
 
@@ -432,41 +521,42 @@ def test_vst(tol=0.02):
     ).T
 
     # Test blind design
-    dds = DeseqDataSet(
-        counts=counts_df, clinical=clinical_df, design_factors=["condition"]
-    )
+    dds = DeseqDataSet(counts=counts_df, metadata=metadata, design_factors=["condition"])
     dds.vst(use_design=False)
     assert (np.abs(r_vst - dds.layers["vst_counts"]) / r_vst).max().max() < tol
 
     # Test full design
-    dds = DeseqDataSet(
-        counts=counts_df, clinical=clinical_df, design_factors=["condition"]
-    )
+    dds = DeseqDataSet(counts=counts_df, metadata=metadata, design_factors=["condition"])
     dds.vst(use_design=True)
     assert (
         np.abs(r_vst_with_design - dds.layers["vst_counts"]) / r_vst_with_design
     ).max().max() < tol
 
 
-def test_ref_level():
+def test_mean_vst(counts_df, metadata, tol=0.02):
+    """
+    Test the output of VST with ``fitType="mean"`` compared with DESeq2.
+    """
+    # Load R data
+    test_path = str(Path(os.path.realpath(tests.__file__)).parent.resolve())
+
+    r_vst = pd.read_csv(
+        os.path.join(test_path, "data/single_factor/r_mean_vst.csv"), index_col=0
+    ).T
+
+    # Test blind design
+    dds = DeseqDataSet(counts=counts_df, metadata=metadata, design_factors=["condition"])
+    dds.vst(use_design=False, fit_type="mean")
+    assert (np.abs(r_vst - dds.layers["vst_counts"]) / r_vst).max().max() < tol
+
+
+def test_ref_level(counts_df, metadata):
     """Test that DeseqDataSet columns are created according to the passed reference
     level, if any.
     """
-    counts_df = load_example_data(
-        modality="raw_counts",
-        dataset="synthetic",
-        debug=False,
-    )
-
-    clinical_df = load_example_data(
-        modality="clinical",
-        dataset="synthetic",
-        debug=False,
-    )
-
     dds = DeseqDataSet(
         counts=counts_df,
-        clinical=clinical_df,
+        metadata=metadata,
         design_factors=["group", "condition"],
         ref_level=["group", "Y"],
     )
@@ -476,7 +566,90 @@ def test_ref_level():
     # Check that its content is correct
     assert (
         dds.obsm["design_matrix"]["group_X_vs_Y"]
-        == clinical_df["group"].apply(
+        == metadata["group"].apply(
             lambda x: 1 if x == "X" else 0 if x == "Y" else np.NaN
         )
     ).all()
+
+
+def test_deseq2_norm(counts_df, metadata):
+    """Test that deseq2_norm() called on a pandas dataframe outputs the same results as
+    DeseqDataSet.fit_size_factors()
+    """
+    # Fit size factors from DeseqDataSet
+    dds = DeseqDataSet(counts=counts_df, metadata=metadata)
+    dds.fit_size_factors()
+    s1 = dds.obsm["size_factors"]
+
+    # Fit size factors from counts directly
+    s2 = deseq2_norm(counts_df)[1]
+
+    np.testing.assert_almost_equal(
+        s1,
+        s2,
+        decimal=8,
+    )
+
+
+@pytest.fixture
+def train_counts(counts_df):
+    return counts_df[25:75]
+
+
+@pytest.fixture
+def test_counts(counts_df):
+    return counts_df[0:25]
+
+
+@pytest.fixture
+def train_metadata(metadata):
+    return metadata[25:75]
+
+
+@pytest.fixture
+def train_dds(train_counts, train_metadata):
+    return DeseqDataSet(
+        counts=train_counts, metadata=train_metadata, design_factors="condition"
+    )
+
+
+def test_deseq2_norm_fit(train_counts):
+    # patient from 25 to 75
+    logmeans, filtered_genes = deseq2_norm_fit(train_counts)
+
+    # 10 genes
+    assert logmeans.shape == (10,)
+    assert filtered_genes.shape == (10,)
+
+
+def test_deseq2_norm_transform(train_counts, test_counts):
+    # First fit with some indices
+    logmeans, filtered_genes = deseq2_norm_fit(train_counts)
+
+    normed_counts, size_factors = deseq2_norm_transform(
+        test_counts, logmeans, filtered_genes
+    )
+    assert isinstance(normed_counts, pd.DataFrame)
+    # 25 samples, 10 genes
+    assert normed_counts.shape == (25, 10)
+    assert size_factors.shape == (25,)
+
+
+def test_vst_fit(train_dds):
+    # patient from 25 to 75
+    train_dds.vst_fit()
+
+    # the correct attributes are fit
+    assert "trend_coeffs" in train_dds.uns
+    assert "normed_counts" in train_dds.layers
+    assert "size_factors" in train_dds.obsm
+
+
+def test_vst_transform(train_dds, test_counts):
+    # First fit with some indices
+    train_dds.vst_fit()
+
+    result = train_dds.vst_transform(test_counts.to_numpy())
+    assert isinstance(result, np.ndarray)
+    # 25 samples, 10 genes
+    assert result.shape == (25, 10)
